
//////////////////////////// RCS COMMENTS: /////////////////////////////
//
//  This is a stub program for the sake of
//  testing the star library.  It performs
//  some operations on the loop data in the
//  star file as if it was an array.  Use this
//  with input redirected from array_usage.str,
//  like so:
//       array_usage < array_usage.str
//
////////////////////////////////////////////////////////////////////////

#include "portability.h"
#include <unistd.h>
#include <stdio.h>
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include "ast.h"
#include "parser.h"

//    PROTOTYPES FOR THE EXAMPLE FUNCTIONS:
//    -------------------------------------
SaveFrameNode * make_summary_frame( DataLoopNode *original_loop_ptr );
SaveFrameNode * make_rowcol_manipulation( DataLoopNode *original_loop_ptr );
SaveFrameNode * make_inner_loop_gone( DataLoopNode *original_loop_ptr );
SaveFrameNode * make_there_can_be_only_one( DataLoopNode *original_loop_ptr );
SaveFrameNode * make_new_nesting_level( DataLoopNode *original_loop_ptr );

// GLOBAL variable: the root of the AST built by a (successful) parse)
StarFileNode *AST;

// output stream:
ofstream * os;

// For printing a string when debugging.  If the C++ string lib
// was not compiled with debug on, then you must call this
// function from the debugger prompt to see the string's contents:
void gdb_str( string const &s )
{
    cout << "len: " << s.length() << " Contents: \"" << s << "\"" << endl;
}

  ///////////////////
 //    main()     //
///////////////////
int main( int argc, char **argv )
{

    // make the star input/output be stdin and stdout:
    // -----------------------------------------------
    os = new ofstream( fileno(stdout) );  // output
    yyin = stdin; // input

    if (yyparse() == 0)   // returns nonzero on error
    {
	// Find the loop containing _outer_looptag_0:
	List<ASTnode*>   *matches;
	List<DataNode*>  *global_list;
	DataLoopNode     *original_loop_ptr;
	GlobalBlockNode  *global_block_ptr;
	SaveFrameNode    *new_frame;

	matches = AST->searchForTypeByTag(
		ASTnode::DATALOOPNODE, string("_outer_looptag_0") );
	
	// IF THIS WERE A REAL LIFE PROGRAM HERE YOU WOULD:
	//     error check if ( matches->size() == 0 )

	// Being sloppy here: just assuming there will only be one match:
	original_loop_ptr = (DataLoopNode*)( (*matches)[0] );
	delete matches;
	
	// Get the global block.  We'll need it later:
	matches = AST->searchForTypeByTag(
		ASTnode::BLOCKNODE, string( "global_" ) );

	// IF THIS WERE A REAL LIFE PROGRAM HERE YOU WOULD:
	//     error check if ( matches->size() == 0 )

	global_block_ptr = (GlobalBlockNode*)( (*matches)[0] );
	delete matches;
	global_list = global_block_ptr->GiveMyDataList();

	// Now do several example operators on it that return new
	// saveframes:

	// Make a summary of info in the outer loop, and add it to the
	// tree in the form of a saveframe.  Note the use of the
	// vector-like "insert" operator to attach it to the global
	// list:
	new_frame = make_summary_frame( original_loop_ptr );
	global_list->insert( global_list->end(), new_frame );

	// Make a version that demonstrates column insertion/removal and
	// row insertion/removal:
	new_frame = make_rowcol_manipulation( original_loop_ptr );
	global_list->insert( global_list->end(), new_frame );

	// Make a version that has the inner loop level removed entirely:
	new_frame = make_inner_loop_gone( original_loop_ptr );
	global_list->insert( global_list->end(), new_frame );
	
	// Make a version that has had all but the first row removed:
	new_frame = make_there_can_be_only_one( original_loop_ptr );
	global_list->insert( global_list->end(), new_frame );

	// Make a version that inserts an entirely new nesting level:
	new_frame = make_new_nesting_level( original_loop_ptr );
	global_list->insert( global_list->end(), new_frame );

	// Now output what we have made:
	AST->Unparse(0);

    }
    else
        cerr << "# Error: Parsing problems in source file"<<endl;


    delete AST;  // yyparse() allocates a tree of data that must be
                 // destructed when finished.  The destructor for
		 // StarFileNode does a complete deep-destruction.

    delete os; // Must do this to close the output, else it never flushes
               // the last line(s) of output.

    return 0;
}
  //////////////////////////
 //  make_summary_frame  //
//////////////////////////
//  Makes a saveframe with the sums of the columns of the loop,
//  and the average (mean) of the columns of the loop:
//
//  Ideas demonstrated in this example:
//      - Iterating over a loop with integer array notation like a table.
//      - Creating a new loop and populating it.
//      - Only the outermost nest level is demonstrated - no inner nests.
//
SaveFrameNode * make_summary_frame( DataLoopNode *original_loop_ptr )
{
    SaveFrameNode         *retVal = new SaveFrameNode(
					    string( "save_summary_frame" ) );
    List<DataNode*>       *sf_insides = retVal->GiveMyDataList();
    LoopTableNode         *table = original_loop_ptr->getValsPtr();
    DataLoopNameListNode  *orig_names = original_loop_ptr->getNamesPtr();

    DataLoopNode          *newLoop;
    DataLoopNameListNode  *nameListList;
    LoopTableNode         *valTable;
    LoopRowNode           *valRow;
    LoopNameListNode      *nameList;

    vector<double>   means, sums;
    int              rowCount, colCount;
    int              curRow, curCol;

    // Give the save frame a comment:
    retVal->setPreComment(
	    string( "# This saveframe contains sums of the info in the\n"
		    "# input loop.  Each row of the loop here has the sum\n"
		    "# and the average (mean) of the column from the\n"
		    "# original loop.\n"
		    "# ----------------------------------------------")  );

    // ------------------------------------------------
    // The first lesson being demonstrated here
    // is how to use the library to parse over
    // the values in the loop input.  To demonstrate
    // this, we will calculate the sums and averages
    // of the columns of numbers in the outermost loop:
    // ------------------------------------------------

    rowCount = table->size();

    // Size of the first row of names (the outermost loop level):
    colCount = (*orig_names)[0]->size();

    // Initialize the sums and means to zero:
    for( curCol = 0 ; curCol < colCount ; curCol++ )
    {
	sums.insert( sums.end(), 0.0 );
	means.insert( means.end(), 0.0 );
    }

    // Calculate sums for each column:
    for(  curRow = 0 ; curRow < rowCount ; curRow++ )
    {   for( curCol = 0 ; curCol < colCount ; curCol++ )
	{
	    // THIS IS THE IMPORTANT LINE FOR THIS EXAMPLE:
	    DataValueNode  *thisVal = (*(*table)[curRow])[curCol] ;
	    sums[curCol] += strtod( thisVal->myValue().c_str(), NULL );
	}
    }

    // Calculate the mean averages for each column:
    for( curCol = 0 ; curCol < colCount ; curCol++ )
	means[curCol] = sums[curCol] / rowCount ;
    
    // --------------------------------------------
    // The second lesson being taught here is how
    // to make a new loop of your own and add
    // things to it.  To demonstrate this, we will
    // make a new loop that shows the stats calced
    // up above:
    // --------------------------------------------

    // Now make the new empty data loop:
    newLoop = new DataLoopNode( string("tabulate") );
    nameListList = newLoop->getNamesPtr();
    valTable = newLoop->getValsPtr();

    // Populate the list of tag names for the loop:
    nameListList->insert( nameListList->end(), new LoopNameListNode() );
    nameList = (*nameListList)[0];
    nameList->insert( nameList->end(), new DataNameNode(
				string( "_original_column_name" ) )  );
    nameList->insert( nameList->end(), new DataNameNode(
				string( "_column_sum" ) )  );
    nameList->insert( nameList->end(), new DataNameNode(
				string( "_column_mean" ) )  );

    // Populate the list of values for the loop, by making new LoopRowNodes
    // and attaching them to the loop's LoopTableNode:
    for( curCol = 0 ; curCol < colCount ; curCol++ )
    {
	char  temp_str[80];

	valRow = new LoopRowNode( true );
	valRow->insert( valRow->end(),
		        new DataValueNode( 
				(*(*orig_names)[0])[curCol]->myName() ,
		        DataValueNode::SINGLE ) );
	sprintf( temp_str, "%lf", sums[curCol] );
	valRow->insert( valRow->end(),new DataValueNode( temp_str )  );
	sprintf( temp_str, "%lf", means[curCol] );
	valRow->insert( valRow->end(), new DataValueNode( temp_str )  );

	valTable->insert( valTable->end(), valRow );
    }

    // Attach the loop we created to the saveframe:
    sf_insides->insert( sf_insides->end(), newLoop );

    // Note that we DON'T delete any of the new'ed objects we created
    // here.  That is because they are now part of the tree.

    return retVal;
}

  ///////////////////////////////////
 //   make_rowcol_manipulation()  //
///////////////////////////////////
SaveFrameNode * make_rowcol_manipulation( DataLoopNode *original_loop_ptr )
{
    SaveFrameNode    *retVal = new SaveFrameNode(
					string("save_rowcol_manipulation") );
    List<DataNode*>  *sf_insides = retVal->GiveMyDataList();
    DataLoopNode     *new_loop_ptr;
    LoopTableNode    *table;

    retVal->setPreComment(
	        string( "# This saveframe demonstrates some insertions\n" 
			"# and deletions of rows and columns.\n"
			"# -------------------------------------------" ) );

    // First, we'll make a copy of the original loop and put it
    // into the new saveframe:
    new_loop_ptr = new DataLoopNode( *original_loop_ptr );
    sf_insides->insert( sf_insides->end(), new_loop_ptr );

    // Now, a ptr to the table of values in the new copy of the loop:
    table = new_loop_ptr->getValsPtr();

    // Okay, for our first manipulation, we'll just remove two rows
    // from the outermost level:
    //
    // WARNING:  THE CODE HERE IS FRAGILE AND WAS WRITTEN UNDER
    // THE KNOWLEGE OF WHAT THE DATA FILE LOOKS LIKE, TO MAKE THIS
    // DEMO PROGRAM SIMPLER TO READ.  IN A REAL LIFE PROGRAM, YOU
    // WILL WANT TO HAVE CHECKS TO MAKE SURE THE VALUES IN QUESTION
    // ACTUALLY EXIST ELSE YOU CAN GET NULL POINTER ERRORS.
    // ------------------------------------------------------------

    //     One way to remove a row is with its destructor.  You should
    //     only delete entire rows this way, not single values.
    //     Safe deletes of single values via destructors is not yet
    //     implemented, but will be eventually.

    delete (*table)[4];  // Remove the fifth row.

    //     Note, because of the above, what used to be subscripts 5,6,7,etc
    //     are now subscript 4,5,6,etc.

    //     Another way to remove a row is with the table's erase operator:
    //     Note how you need to pass not the pointer to the row, but a
    //     pointer to the pointer to the row.  That's because the table holds
    //     a list of POINTERS to LoopRowNode's, and the vector operators
    //     want to be passed a pointer to what is stored in them.

    table->erase( &((*table)[4]) ); // Remove the new fifth row (which was
	                            // once the sixth row when we started).


    // Oh, note that one of the rows above had inner loop values under it
    // in the original loop.  They are gone too when it is removed.

    // For the next statement, we'll remove the third (2th) row from the
    // inner loop that is under the first (zeroth) row of the outer
    // loop (It's the "inval2  inval9" row that we are deleting here):
    // -----------------------------------------------------------------

    (*table)[0]->myLoop()->erase( &( (*((*table)[0])->myLoop() )[2] ) );

    // For the next statement, we'll remove a column.  To remove a
    // column, you just remove it's tag name from the name list, and
    // it will remove all the values that go with it:
    // -------------------------------------------------------------

    //     This removes the first (zeroth) column from the OUTER loop:

    new_loop_ptr->getNames()[0]->erase(
	    & ( ( *(new_loop_ptr->getNames()[0]) )[0] ) );

    //     This removes the first (zeroth) column from the INNER loop:
    //     (The index to getNames is the nesting level of the tag name).

    new_loop_ptr->getNames()[1]->erase(
	    & ( *(new_loop_ptr->getNames()[1]) )[0] );

    //     Note that this would have also worked as the argument to erase():
    //          ( *(new_loop_ptr->begin()) )->begin()
    //     In the above line.  begin() returns an iterator of the start
    //     of the list, which is exactly what you get when you take the
    //     address of the thing at subscript zero:

    // Okay, enough removing.  We are now left with a loop that has had
    // the 4th and 5th outer rows removed, and the last row of the
    // loop under the first row removed, and has had the first columns
    // removed from both the inner and outer loop.  We are left with a
    // loop that has 2 outer columns, and 1 inner column.  Lets insert
    // some things into that loop:
    // ----------------------------------------------------------------

    //     These lines make a new loop row and inserts it to the outermost
    //     level, at the end:

    LoopRowNode   *newRow;

    newRow = new LoopRowNode( true );
    newRow->insert( newRow->end(), new DataValueNode("9001") );
    newRow->insert( newRow->end(), new DataValueNode("9002") );
    table->insert( table->end(), newRow );

    //     These lines do the same thing, but insert at the front:
    newRow = new LoopRowNode( true) ;
    newRow->insert( newRow->end(), new DataValueNode("1001") );
    newRow->insert( newRow->end(), new DataValueNode("1002") );
    table->insert( table->begin(), newRow );

    //     These lines do the same thing, but insert before the 3rd row:
    newRow = new LoopRowNode( true );
    newRow->insert( newRow->end(), new DataValueNode("5001") );
    newRow->insert( newRow->end(), new DataValueNode("5002") );
    table->insert( &((*table)[3]), newRow );
    
    //     These lines are an example of an invalid insertion.  The
    //     outermost loop has exactly 2 columns (remember we deleted one
    //     up above), and therefore you cannot insert a row with
    //     less than or more than exactly 2 values at that level:
    newRow = new LoopRowNode( true );
    newRow->insert( newRow->end(),
	            new DataValueNode("dummyval",DataValueNode::SINGLE) );
    if(  ! table->insert( table->end(), newRow )  )
    {
	delete newRow; // delete it since it did not get put into the
	               // star tree.
	// TODO: You could have error handling here.
    }

    // Now, let's try to insert a column.  We'll insert it at the
    // innermost level:  Remember that at this point the loop now
    // has 2 columns at the outer level (level 0) and 1 column at
    // the inner level (level 1).  We'll add two columns to the inner
    // level, making it a structure of 2 and then 3.
    // We can pass a dummy new value as the value to fill the columns
    // with (inserting a new name causes it to create new values in the
    // rows where it needs them to remain valid):
    // --------------------------------------------------------------
   
    DataLoopNameListNode  *names = new_loop_ptr->getNamesPtr();
    // (*names)[1] is the inner loop.  (*names)[0] would be the outer loop).

    //     This example lets the library insert the default value (.) where
    //     needed:
    (*names)[1]->insert( (*names)[1]->begin(),
			 new DataNameNode( "_new_first_inner_tag" ) );
    //     This example specifies the value to insert where needed:
    (*names)[1]->insert( (*names)[1]->end(),
			 new DataNameNode( "_new_last_inner_tag" ),
			 DataValueNode( "FillerVal" )  );

    // Now, lastly, let's insert a new inner row.  Keep in mind that now
    // all the inner rows have to have three columns because of the
    // two we just added above:
    // -----------------------------------------------------------------

    newRow = new LoopRowNode( true );
    newRow->insert( newRow->end(), new DataValueNode( "new_inner_val1" ) );
    newRow->insert( newRow->end(), new DataValueNode( "new_inner_val2" ) );
    newRow->insert( newRow->end(), new DataValueNode( "new_inner_val3" ) );
    (*table)[1]->myLoop()->insert( (*table)[1]->myLoop()->begin(), newRow );


    // Okay, that's enough examples.  Any more and it would get hard
    // to follow the changes when looking at the output saveframe.

    return retVal;
}

SaveFrameNode * make_inner_loop_gone( DataLoopNode *original_loop_ptr )
{
    SaveFrameNode    *retVal = new SaveFrameNode(
					string("save_inner_loop_gone") );
    List<DataNode*>  *sf_insides = retVal->GiveMyDataList();
    DataLoopNode     *new_loop_ptr;

    retVal->setPreComment(
	    string( "# This saveframe has deleted the inner loop\n"
		    "# altogether by removing the inner tag names.\n"
		    "# -------------------------------------------\n" ) );

    // First, we'll make a copy of the original loop and put it
    // into the new saveframe:
    // ---------------------------------------------------------
    new_loop_ptr = new DataLoopNode( *original_loop_ptr );
    sf_insides->insert( sf_insides->end(), new_loop_ptr );
    
    // Then, we eliminate both tags from the inner loop.
    // (Notice that after eliminating the zeroth inner tag, that
    // what was once the one-th inner tag has become the zeroth tag.
    // That is why we delete the "zeroth tag" twice.)
    // -------------------------------------------------------------
    new_loop_ptr->getNames()[1]->erase(
	    &( ( *(new_loop_ptr->getNames()[1]) )[0] )  );
    new_loop_ptr->getNames()[1]->erase(
	    &( ( *(new_loop_ptr->getNames()[1]) )[0] )  );

    // Now that both inner tags are gone, the library automatically
    // deletes all the inner loops from the loop node.  When we
    // print this out, all we get is the outermost loop:

    return retVal;
}

SaveFrameNode * make_there_can_be_only_one( DataLoopNode *original_loop_ptr )
{
    SaveFrameNode    *retVal = new SaveFrameNode(
	    				string("save_there_can_be_only_one") );
    List<DataNode*>  *sf_insides = retVal->GiveMyDataList();
    LoopTableNode    *table = original_loop_ptr->getValsPtr();

    retVal->setPreComment( string( "# This saveframe is not implemented.\n" ) );

    return retVal;
}

SaveFrameNode * make_new_nesting_level( DataLoopNode *original_loop_ptr )
{
    SaveFrameNode    *retVal = new SaveFrameNode(
	    				string("save_new_nesting_level") );
    List<DataNode*>  *sf_insides = retVal->GiveMyDataList();
    LoopTableNode    *table = original_loop_ptr->getValsPtr();

    retVal->setPreComment( string( "# This saveframe is not implemented.\n" ) );

    return retVal;
}
